'''
Created on 2013-3-28

    pysa - reverse a complete computer setup
    Copyright (C) 2013  MadeiraCloud Ltd.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Ken
'''
#------------------------------------------------------------
import logging
import time

from pysa.scanner.actions.file import scanner_file
from pysa.scanner.actions.gem import scanner_gem
from pysa.scanner.actions.npm import scanner_npm
from pysa.scanner.actions.php import scanner_php
from pysa.scanner.actions.pypi import scanner_pypi
from pysa.scanner.actions.service import scanner_service
from pysa.scanner.actions.host import scanner_host
from pysa.scanner.actions.mount import scanner_mount
from pysa.scanner.actions.cron import scanner_cron
from pysa.scanner.actions.sshkey import scanner_key
from pysa.scanner.actions.user import scanner_user
from pysa.scanner.actions.group import scanner_group
from pysa.scanner.actions.package import scanner_package
from pysa.scanner.actions.source import scanner_source
from pysa.scanner.actions.repository import scanner_repo
from pysa.scanner.actions.process import scanner_process
from pysa.scanner.actions.base import scanner_base
#------------------------------------------------------------

class scanner_handler():

    # stay aware of the order
    handler = {
                "file"      : scanner_file,
                "gem"       : scanner_gem,
                "npm"       : scanner_npm,
                "php"       : scanner_php,
                "pypi"      : scanner_pypi,
                "service"   : scanner_service,
                "host"      : scanner_host,
                'user'      : scanner_user,
                'group'     : scanner_group,
                'mount'     : scanner_mount,
                'cron'      : scanner_cron,
                'key'       : scanner_key,
                'package'   : scanner_package,
                'source'    : scanner_source,
                'repository': scanner_repo,
                'process'   : scanner_process
               }



    def __init__(self, rules):
        self.resources  =   {
                                'packages'  :   {},
                                'files'     :   {},
                                'crons'     :   {},
                                'groups'    :   {},
                                'mounts'    :   {},
                                'hosts'     :   {},
                                'repos'     :   {},
                                'services'  :   {},
                                'keys'      :   {},
                                'users'     :   {},
                                'ips'       :   [],
                                'sources'   :   {},
                                'proces'    :   {}

                             }

        self.rules = rules if rules else {}

    def scan(self):

        # init the base scanner
        s = scanner_base(
                self.resources['packages'],
                self.resources['files'],
                self.resources['crons'],
                self.resources['groups'],
                self.resources['mounts'],
                self.resources['hosts'],
                self.resources['repos'],
                self.resources['services'],
                self.resources['keys'],
                self.resources['users'],
                self.resources['ips'],
                self.resources['sources'],
                self.resources['proces']
        )
        # init the filter rules
        s.init_filter(self.rules)

        for scanner_key in self.handler.keys():
            # ignore the discard resources
            if 'discard' in self.rules and '_resources' in self.rules['discard'] and scanner_key in self.rules['discard']['_resources']: continue

            # time begin
            time_begin = time.time()

            # log
            logging.info('ScannerHandler.scan(): Scanning Module %s, time begin at %s ' % (scanner_key, time_begin))

            # scan according to different modules
            try:
                s.__class__ = self.handler[scanner_key]
                
                s.scan()

            except Exception, e:
                logging.error('ScannerHandler.scan(): %s Error message, %s' % (scanner_key,str(e)))

            # time end
            time_consume = time.time() - time_begin
            logging.info('ScannerHandler.scan(): Scanning Module %s, time consume %s ' % (scanner_key, time_consume))

        return self.resources

def module_scan(filters = None):
    scanner = scanner_handler(filters)
    return scanner.scan()



'''
Generate puppet scripts

    pysa - reverse a complete computer setup
    Copyright (C) 2013  MadeiraCloud Ltd.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Thibault BRONCHAIN
'''

import re

from pysa.exception import *
from pysa.config import *
from pysa.tools import *
from pysa.output import output

from pysa.puppet.puppet_converter import GLOBAL_SEC_EQ

# define quoted variables
QUOTED_AVOIDED_KEYS = ['content', 'before']
QUOTED_KEYS = ['checksum', 'name', 'group', 'owner']
QUOTED_CONTENT = ['\W', '\d']

# puppet generation class
class puppet_build():
    def __init__(self, class_dict, output_path, module):
        self.__prepross = [
            [GLOBAL_SEC_EQ['Exec'], self.__content],
            ]

        self.__quoted_regex = "%s" % ('|'.join(QUOTED_CONTENT))
        self.__module_name = module
        self.__c = output()
        self.__class_dict = class_dict
        self.__output_path = output_path+'/'+self.__module_name
        self.__prepross_list = []
        for i in self.__prepross:
            self.__prepross_list.append(i[0])

    # main function
    @general_exception
    def run(self):
        tools.l(INFO, "running generation engine", 'run', self)
        self.__generate(self.__class_dict)
        self.__init_gen()
        self.dump_in_files()
        tools.l(INFO, "complete.", 'run', self)
        return True

    # print the puppet files
    @general_exception
    def dump(self):
        for c in self.__c.list():
            print "%s:\n%s\n\n" % ((c if c else 'init'),self.__c.dump(c))

    # dump puppet file in variable
    @general_exception
    def dump_in_var(self, data=''):
        for c in self.__c.list():
            data += ("%s:\n%s\n\n" % ((c if c else 'init'),self.__c.dump(c)))
        return data

    # dump the puppet files into the right files
    @general_exception
    def dump_in_files(self):
        for c in self.__c.list():
            tools.write_in_file(self.__output_path+'/manifests/'+(c if c else 'init')+'.pp', self.__c.dump(c))

    # init file generation
    @general_exception
    def __init_gen(self):
        inc=''
        for c in self.__c.list():
            inc += ("include %s\n" % (c) if c else '')
        tab = re.split('\n', self.__c.dump()+'\n'+inc)
        s=''
        for i in tab:
            s += (re.sub(r'^', r'\n\t', i) if i else '')
        self.__c.mod("class %s {\n%s\n}\n" % (self.__module_name, s))

    # particular case for the single instructions
    @general_exception
    def __single_instruction(self, parent, data, index, output, tab):
        if not parent:
            return tab
        for content in data[index]:
            if index == GLOBAL_SEC_EQ['require']:
                if content not in parent:
                    continue
            self.__c.add(output, tab+"%s %s\n" % (index[len(SINGLE_EQ):],content))
        return tab

    # quote required values
    @general_exception
    def __quote_values(self, key, val):
        return (("'%s'" % (re.sub('\'', '\\\'', val))
                 if (key not in QUOTED_AVOIDED_KEYS)
                 and ((key in QUOTED_KEYS)
                      or (re.search(self.__quoted_regex, val)))
                 else val)
                if type(val) is str else val)

    # content writing
    @general_exception
    def __content_writing(self, output, sectype, label, optlabel, content):
        i = 0
        s = ''
        if (type(content) is list):
            tmp_s = ''
            for sec in content:
                tmp_s += (", " if i else '')+"'%s'" % (sec)
                i += 1
            if i > 0:
                s += ("[" if i > 1 else '')+tmp_s+("]" if i > 1 else '')
            else: s = None
        elif (type(content) is dict):
            tmp_s = ''
            for sec in content:
                for d in content[sec]:
                    tmp_s += (", " if i else '') + "%s['%s']" % (sec,d)
                    i += 1
            if i > 0:
                s += ("[" if i > 1 else '')+tmp_s+("]" if i > 1 else '')
            else: s = None
        else:
            if (sectype == "_file") and (label[0] != '-') and optlabel == 'content':
                tools.write_in_file(self.__output_path+'/templates'+('/' if label[0] != '/' else '')+label,
                                    content)
                content = ("template('%s')" % (self.__module_name+('/' if label[0] != '/' else '')+label))
            s += "%s" % self.__quote_values(optlabel, content)
        return s

    # global content generation for pupept config file
    @general_exception
    def __content(self, parent, data, index, output, tab):
        tools.l(INFO, "creating section %s" % (index.lstrip(VOID_EQ)), 'content', self)
        if index[:len(SINGLE_EQ)] == SINGLE_EQ:
            return self.__single_instruction(parent, data, index, output, tab)
        self.__c.add(output, "%s%s {\n" % (tab,index.lstrip(VOID_EQ)))
        for label in sorted(data[index]):
            if label in NULL:
                continue
            elif label[0] != VOID_EQ:
                tab = tools.tab_inc(tab)
                self.__c.add(output, "%s'%s':\n" % (tab,label))
            f = 0
            tab = tools.tab_inc(tab)
            for optlabel in sorted(data[index][label]):
                if (data[index][label][optlabel] not in NULL) and (optlabel[0] != VOID_EQ):
                    s = self.__content_writing(output,
                                               index,
                                               label,
                                               optlabel,
                                               data[index][label][optlabel])
                    if s:
                        self.__c.add(output, (",\n" if f else '')+
                                     "%s%s => %s" % (tab, optlabel,s))
                        f = 1
            if f and label != VOID_EQ:
                self.__c.add(output, ";\n")
            elif f and label == VOID_EQ:
                self.__c.add(output, "\n")
            tab = tools.tab_dec(tab)
            if label[0] != VOID_EQ:
                tab = tools.tab_dec(tab)
        self.__c.add(output, "%s}\n" % (tab))
        return tab

    # class generation method, applies the recursion
    @general_exception
    def __object(self, parent, data, index, output, tab, level):
        tools.l(INFO, "generation class %s" % (index), 'object', self)
        self.__c.add(output, "%sclass %s {\n" % (tab,index))
        # recursion here
        tab = tools.tab_dec(self.__generate(data[index],
                                            data,
                                            output,
                                            tools.tab_inc(tab),
                                            level+1))
        self.__c.add(output, "%s}\n%sinclude %s\n" % (tab,tab,index))
        return tab

    # puppet file generation function
    # recursive function
    @general_exception
    def __generate(self, data, parent = None, output='', tab='', level=0):
        # preprocessing exceptions
        for t in self.__prepross:
            if t[0] in data and t[1]:
                tools.l(INFO, "adding %s condition" % (t[0]), 'generate', self)
                tab = t[1](parent, data, t[0], output, tab)
        # global generation
        for index in sorted(data):
            # particular cases needed to be at the beginning of a section
            if index in self.__prepross_list:
                continue
            # content found
            elif index[0] == VOID_EQ and output:
                tab = self.__content(parent, data, index, output, tab)
            # new class
            else:
                if (level == 0):
                    output = index
                    self.__c.add_dict(output)
                # recursion here
                tab = self.__object(parent, data, index, output, tab, level)
        return tab



'''
List dependencies for all resources

    pysa - reverse a complete computer setup
    Copyright (C) 2013  MadeiraCloud Ltd.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Thibault BRONCHAIN
'''

import re

from pysa.tools import *
from pysa.config import *
from pysa.exception import *

from pysa.puppet.puppet_converter import SECTION_EQ

SCM_EQ = {
    'git' : ['git', 'git-all'],
    'svn' : ['subversion'],
    'hg'  : ['mercurial'],
}

SELF_ORDER = {
    'dirs'          : 'path',
    'sources'       : 'path',
    'mounts'        : 'name',
    }

PRIOR = ['sources']

class dependencies:
    def __init__(self, data=None):
        # define dependencies
        self.__deps = {
            'hosts'     : {},
            'mounts'    : {
                'hosts'     : None,
                'mounts'    : [
                    self.__based_on_field, {
                        'field' : [self.__self_order, {
                                'field'     : 'name',
                                }],
                        'key'   : 'name',
                        }
                    ],
                },
            'groups'    : {
                'hosts'     : None,
                },
            'users'     : {
                'hosts'     : None,
                'groups'    : [
                    self.__based_on_field, {
                        'field' : 'group',
                        'key'   : 'name',
                        }
                    ],
                },
            'dirs'      : {
                'hosts'     : None,
                'dirs'    : [
                    self.__based_on_field, {
                        'field' : [self.__self_order, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                'groups'    : [
                    self.__based_on_field, {
                        'field' : 'group',
                        'key'   : 'name',
                        }
                    ],
                'users'     : [
                    self.__based_on_field, {
                        'field' : 'owner',
                        'key'   : 'name',
                        }
                    ],
                'mounts'    : [
                    self.__based_on_field, {
                        'field' : [self.__get_mount_from_path, {
                                'field'     : 'path',
                                }],
                        'key'   : 'device',
                        }
                    ],
                },
            'keys'      : {
                'hosts'     : None,
                'groups'    : [
                    self.__based_on_field, {
                        'field' : 'group',
                        'key'   : 'name',
                        }
                    ],
                'users'     : [
                    self.__based_on_field, {
                        'field' : 'owner',
                        'key'   : 'name',
                        }
                    ],
                'dirs'      : [
                    self.__based_on_field, {
                        'field' : [self.__get_base_path, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                },
            'repos'     : {
                'hosts'     : None,
                'groups'    : [
                    self.__based_on_field, {
                        'field' : 'group',
                        'key'   : 'name',
                        }
                    ],
                'users'     : [
                    self.__based_on_field, {
                        'field' : 'owner',
                        'key'   : 'name',
                        }
                    ],
                'dirs'      : [
                    self.__based_on_field, {
                        'field' : [self.__get_base_path, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                },
            'packages'  : {
                'hosts'     : None,
                'repos'     : [
                    self.__based_on_field, {
                        'field' : 'provider',
                        'key'   : 'provider',
                        }
                    ],
                'dirs'      : [
                    self.__based_on_field, {
                        'field' : [self.__get_base_path, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                },
            'files'     : {
                'hosts'     : None,
                'groups'    : [
                    self.__based_on_field, {
                        'field' : 'group',
                        'key'   : 'name',
                        }
                    ],
                'users'     : [
                    self.__based_on_field, {
                        'field' : 'owner',
                        'key'   : 'name',
                        }
                    ],
                'dirs'      : [
                    self.__based_on_field, {
                        'field' : [self.__get_base_path, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                'packages'  : [
                    self.__based_on_field, {
                        'field' : 'path',
                        'key'   : 'config_files',
                        }
                    ],
                },
            'crons'     : {
                'hosts'     : None,
                'users'     : [
                    self.__based_on_field, {
                        'field' : 'user',
                        'key'   : 'name',
                        }
                    ],
                },
            'sources'   : {
                'hosts'     : None,
                'sources'   : [
                    self.__based_on_field, {
                        'field' : [self.__self_order, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                'groups'    : [
                    self.__based_on_field, {
                        'field' : 'group',
                        'key'   : 'name',
                        }
                    ],
                'users'     : [
                    self.__based_on_field, {
                        'field' : 'owner',
                        'key'   : 'name',
                        }
                    ],
                'dirs'      : [
                    self.__based_on_field, {
                        'field' : [self.__get_base_path, {
                                'field'     : 'path',
                                }],
                        'key'   : 'path',
                        }
                    ],
                'keys'      : [
                    self.__based_on_field, {
                        'field' : 'key',
                        'key'   : 'name',
                        }
                    ],
                'packages'  : [
                    self.__based_on_field, {
                        'field' : [self.__get_pkg_from_scm, {
                                'field'     : 'scm',
                                }],
                        'key'   : 'name',
                        }
                    ],
                },
            'services'  : {
                'files'     : None,
                },
            }

        self.__data = data


    @general_exception
    def run(self, data = None):
        tools.l(INFO, "running dependency cycle generation", 'run', self)
        if data:
            self.__data = data
        for c in self.__data:
            if c not in self.__deps: continue
            for obj_name in self.__data[c]:
                obj = self.__data[c][obj_name]
                for dep_name in PRIOR:
                    if dep_name in self.__deps[c]:
                        self.__parse_dep(c, obj_name, obj, dep_name)
                for dep_name in self.__deps[c]:
                    if dep_name in PRIOR: continue
                    else: self.__parse_dep(c, obj_name, obj, dep_name)
        tools.l(INFO, "dependency cycle generated", 'run', self)

    @general_exception
    def __parse_dep(self, c, obj_name, obj, dep_name):
        dep = self.__deps[c][dep_name]
        if not dep:
            obj['require'] = tools.dict_merging(obj.get('require'), {
                    'Class' : [dep_name]
                    })
        elif type(dep) is list:
            res = dep[0](obj, dep_name, dep[1])
            if res:
                obj['require'] = tools.dict_merging(obj.get('require'), {
                        SECTION_EQ[dep_name][len(VOID_EQ):].capitalize() : res
                        })

    @general_exception
    def __self_order(self, object, gclass, args):
        ref = object.get(args['field'])
        data = self.__data[gclass]
        if not data: return None
        res = None
        for key in sorted(data, key=lambda x: data[x][args['field']]):
            name = data[key][args['field']]
            if re.match(name, ref) and (ref != name):
                res = key
        if gclass == 'sources' and self.__data.get('dirs') and res:
            dirs = dict(self.__data['dirs'].items())
            for dir in dirs:
                comp = tools.path_basename(ref)
                if dirs[dir]['path'] == comp:
                    self.__data['dirs'].pop(dir)
        return res

    @general_exception
    def __get_mount_from_path(self, object, gclass, args):
        path = object.get(args['field'])
        mounts = self.__data['mounts']
        if not mounts: return None
        res = None
        for key in sorted(mounts, key=lambda x: mounts[x]['name']):
            name = mounts[key]['name']
            if re.match(name, path):
                res = key
        return res

    @general_exception
    def __get_pkg_from_scm(self, object, gclass, args):
        scm = object.get(args['field'])
        return (SCM_EQ.get(scm) if scm else None)

    @general_exception
    def __get_base_path(self, object, gclass, args):
        path = object.get(args['field'])
        return (tools.path_basename(path) if path else None)

    @general_exception
    def __based_on_field(self, object, gclass, args):
        if not args.get('field') or not args.get('key'):
            return []
        res = []
        v_field = (args['field'][0](object, gclass, args['field'][1])
                   if type(args['field']) is list
                   else object.get(args['field']))
        if v_field:
            for obj_name in self.__data[gclass]:
                v_key = (args['key'][0](object, gclass, obj_name, args['key'][1])
                         if type(args['key']) is list
                         else self.__data[gclass][obj_name].get(args['key']))
                if (((type(v_field) is list) and (v_key in v_field))
                    or ((type(v_key) is list) and (v_field in v_key))
                    or (v_key == v_field)):
                    res.append(obj_name)
        return res



'''
Puppet Objects

    pysa - reverse a complete computer setup
    Copyright (C) 2013  MadeiraCloud Ltd.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Thibault BRONCHAIN
'''

from pysa.exception import *

from pysa.scanner.actions.utils import get_stat

class puppet_objects():
    @staticmethod
    def puppet_file_dir_obj(dr):
        # get group, mode and owner
        s = get_stat(dr)
        #s = ('root', oct(0777), 'root')
        return {
            'path'      : dr,
            'ensure'    : 'directory',
            'name'      : dr,
            'group'     : s[0],
            'mode'      : s[1],
            'owner'     : s[2],
            }

PUPPET_OBJ_MAKER = {
    'file' : puppet_objects.puppet_file_dir_obj
    }
